import os, glob, subprocess, sys
import numpy as np
import pandas as pd
from functools import partial
from scipy.optimize import basinhopping
from multiprocessing import Pool, Manager
from sklearn.metrics import roc_curve, auc
from libraries import mlayerlinkpred as mlp

def fit_ensemble(df, methods, x):
   df["ensemble"] = 0
   for i in range(len(methods)):
      df["ensemble"] += (x[i] * df[methods[i]])
   fpr, tpr, thresholds = roc_curve(df["target"], df["ensemble"])
   return 1 - auc(fpr, tpr)

def _moss_wrapper(x):
   s = {
      "aarhus": 20,
      "phys": 15,
      "celegans": 75
   }
   d = "-D" if x[0] != "aarhus" else ""
   bash_command = f"java -Xmx16g -cp libraries/moss_new.jar moss.Miner -inel -onel -x -m2 -n4 -s{s[x[0]]} -k3 {d} -Zdata/moss{x[1]}.rul -w0 -T data/{x[0]}/{x[0]}{x[1]}.gspan data/moss{x[1]}.sub"
   process = subprocess.Popen(bash_command.split(), stdout = subprocess.PIPE, stderr = subprocess.PIPE)
   output, error = process.communicate()
   bash_command = f"java -Xmx16g -cp libraries/moss_new.jar moss.Predictor -fnel {d} data/moss{x[1]}.sub data/moss{x[1]}.rul data/{x[0]}/{x[0]}{x[1]}.gspan data/moss{x[1]}.ped"
   process = subprocess.Popen(bash_command.split(), stdout = subprocess.PIPE, stderr = subprocess.PIPE)
   output, error = process.communicate()

def moss_wrapper(network):
   pool = Pool(processes = nproc[network])
   _ = pool.map(_moss_wrapper, [(network, fold) for fold in range(10)])
   pool.close()
   pool.join()
   dfs = []
   for txt in glob.glob("data/moss*.ped"):
      netid = txt.split("moss")[1].split('.')[0].split('_')[0]
      df = pd.read_csv(txt, sep = "\t")
      df = df.rename(columns = {"dst": "trg", "edge": "layer", "wgt": "count", "lift1": "lift"})
      df.loc[df["src"] == 0, "src"] = df.loc[df["src"] == 0, "node"].map(lambda x : f"new{x:.0f}")
      df.loc[df["trg"] == 0, "trg"] = df.loc[df["trg"] == 0, "node"].map(lambda x : f"new{x:.0f}")
      df["fold"] = int(netid)
      dfs.append(df.drop(["graph", "node", "lift2", "count", "lift"], axis = 1))
   for f in glob.glob(f"data/moss*"):
      os.remove(f)
   df = pd.concat(dfs)
   df = df.groupby(by = ["src", "trg", "layer", "fold"]).sum().reset_index().rename(columns = {"conf": "new"})
   df["src"] = df["src"].astype(str)
   df["trg"] = df["trg"].astype(str)
   df["layer"] = df["layer"].astype(int)
   return df

def _magma_wrapper(x):
   s = {
      "aarhus": 50,
      "phys": 45,
      "celegans": 85
   }
   mlp.magma(f"data/{x[0]}/{x[0]}{x[1]}", s[x[0]], 4, directed = x[0] != "aarhus")

def magma_wrapper(network):
   pool = Pool(processes = nproc[network])
   _ = pool.map(_magma_wrapper, [[network, fold] for fold in range(10)])
   pool.close()
   pool.join()
   dfs = []
   for txt in glob.glob(f"data/{network}/*magmascore*"):
      netid = txt.split("_")[0].split(network)[-1]
      df = pd.read_csv(txt, sep = "\t")
      df["fold"] = int(netid)
      df.loc[df["src"] < 0, "src"] = df.loc[df["src"] < 0, "src"].map(lambda x: f"new{x*-1}")
      df.loc[df["trg"] < 0, "trg"] = df.loc[df["trg"] < 0, "trg"].map(lambda x: f"new{x*-1}")
      dfs.append(df)
      os.remove(txt)
   for f in glob.glob(f"data/{network}/*.rules"):
      os.remove(f)
   for f in glob.glob(f"data/{network}/*.mossout"):
      os.remove(f)
   df = pd.concat(dfs)
   df = df.groupby(by = ["src", "trg", "layer", "fold"]).sum().reset_index()
   df["src"] = df["src"].astype(str)
   df["trg"] = df["trg"].astype(str)
   df["layer"] = df["layer"].astype(int)
   return df.rename(columns = {"magma": "old"})

def traditional_wrapper(x):
   directed = x[0] != "aarhus"
   fold_df = mlp.sharma(f"data/{x[0]}/{x[0]}{x[1]}", directed = directed)
   fold_df = fold_df.merge(mlp.pujari(f"data/{x[0]}/{x[0]}{x[1]}", directed = directed), on = ["src", "trg", "layer"], how = "outer")
   fold_df = fold_df.merge(mlp.jalili(f"data/{x[0]}/{x[0]}{x[1]}", directed = directed), on = ["src", "trg", "layer"], how = "outer")
   fold_df = fold_df.merge(mlp.hristova(f"data/{x[0]}/{x[0]}{x[1]}", directed = directed), on = ["src", "trg", "layer"], how = "outer")
   fold_df = fold_df.merge(mlp.debacco(f"data/{x[0]}/{x[0]}{x[1]}", directed = directed), on = ["src", "trg", "layer"], how = "outer")
   fold_df["fold"] = x[1]
   x[2].append(fold_df.fillna(0))

def mell_wrapper(network):
   results = []
   for fold in range(10):
      results.append(mlp.mell(f"data/{network}/{network}{fold}", directed = network != "aarhus"))
      results[fold]["fold"] = fold
   return pd.concat(results)

def ensemble(scores, network):
   methods = ["debacco", "hristova", "jalili", "old", "new", "pujari", "sharma", "mell"]
   for lp in methods:
      scores[lp] = (scores[lp] - scores[lp].mean()) / scores[lp].std()
   scores["ensemble_base"] = scores[methods].mean(axis = 1)
   df = test_set(network).merge(scores, on = ["src", "trg", "layer", "fold"], how = "left")
   for lp in methods:
      df[lp] = df[lp].fillna(df[lp].min())
   x0 = [1.0] * len(methods)
   res = basinhopping(partial(fit_ensemble, df, methods), x0)
   scores["ensemble_optim"] = 0
   for i in range(len(methods)):
      scores["ensemble_optim"] += (res.x[i] * scores[methods[i]])

def run_mlps(network):
   sys.stderr.write(f"Running all traditiona LPs on {network} in parallel...\n")
   manager = Manager()
   results = manager.list()
   pool = Pool(processes = nproc[network])
   _ = pool.map(traditional_wrapper, [[network, fold, results] for fold in range(10)])
   pool.close()
   pool.join()
   results = pd.concat(list(results))
   sys.stderr.write(f"Running MELL on {network}...\n")
   results = results.merge(mell_wrapper(network), on = ["src", "trg", "layer", "fold"], how = "outer").fillna(0)
   results["src"] = results["src"].astype(str)
   results["trg"] = results["trg"].astype(str)
   sys.stderr.write(f"Running old code on {network}...\n")
   results = results.merge(magma_wrapper(network), on = ["src", "trg", "layer", "fold"], how = "outer").fillna(0)
   sys.stderr.write(f"Running new code on {network}...\n")
   results = results.merge(moss_wrapper(network), on = ["src", "trg", "layer", "fold"], how = "outer").fillna(0)
   sys.stderr.write(f"Running ensemble on {network}...\n")
   ensemble(results, network)
   return results

def test_set(network):
   directed = network != "aarhus"
   tests = []
   for fold in range(10):
      target = pd.read_csv(f"data/{network}/{network}{fold}_test", sep = "\t", header = None, names = ("src", "trg", "layer", "fold"), dtype = {"src": str, "trg": str})
      df2 = pd.read_csv(f"data/{network}/{network}{fold}", sep = "\t", header = None, names = ("src", "trg", "layer"), dtype = {"src": str, "trg": str})
      df2["fold"] = fold
      srcs = set(df2["src"])
      trgs = set(df2["trg"])
      layers = set(df2["layer"])
      if directed:
         edges = [(src, trg, layer, fold) for src in srcs for trg in trgs for layer in layers if src != trg]
      else:
         edges = [(src, trg, layer, fold) for src in srcs for trg in trgs for layer in layers if int(src) < int(trg)]
      test = pd.DataFrame(data = edges, columns = ("src", "trg", "layer", "fold"))
      test = pd.concat([test, target, df2])
      test = test.drop_duplicates(keep = False)
      test = test.sample(target.shape[0])
      test["target"] = 0
      target["target"] = 1
      tests.append(pd.concat([test, target]))
   return pd.concat(tests)

def evaluate(scores, test):
   test = test.merge(scores, on = ["src", "trg", "layer", "fold"], how = "left")
   evaluation = []
   for lp in ("debacco", "hristova", "jalili", "pujari", "sharma", "mell", "old", "new", "ensemble_base", "ensemble_optim"):
      test[lp] = test[lp].fillna(test[lp].min())
      fpr, tpr, thresholds = roc_curve(test["target"], test[lp])
      evaluation.append((lp, auc(fpr, tpr), fpr, tpr))
   evaluation = pd.DataFrame(data = evaluation, columns = ("method", "auc", "fpr", "tpr"))
   evaluation = evaluation.set_index("method")
   btc = evaluation.loc[["debacco", "hristova", "jalili", "pujari", "sharma"]].sort_values(by = "auc").index[-1]
   np.savetxt(f"fig_12{figlabel[i]}_red.csv", np.vstack((evaluation.loc["old", "fpr"], evaluation.loc["old", "tpr"])).T, delimiter = "\t", fmt = "%.8f")
   np.savetxt(f"fig_12{figlabel[i]}_green.csv", np.vstack((evaluation.loc["new", "fpr"], evaluation.loc["new", "tpr"])).T, delimiter = "\t", fmt = "%.8f")
   np.savetxt(f"fig_12{figlabel[i]}_blue.csv", np.vstack((evaluation.loc["mell", "fpr"], evaluation.loc["mell", "tpr"])).T, delimiter = "\t", fmt = "%.8f")
   np.savetxt(f"fig_12{figlabel[i]}_purple.csv", np.vstack((evaluation.loc[btc, "fpr"], evaluation.loc[btc, "tpr"])).T, delimiter = "\t", fmt = "%.8f")
   evaluation["auc"].reset_index().to_csv(f"tab2_col{i + 1}.csv", index = False, sep = "\t")

nproc = {
  "aarhus": 10,
  "phys": 10,
  "celegans": 2
}

datasets = ["aarhus", "phys", "celegans"]
figlabel = "abc"

for i in range(len(datasets)):
   scores = run_mlps(datasets[i])
   sys.stderr.write(f"Building test set of {datasets[i]}...\n")
   test = test_set(datasets[i])
   sys.stderr.write(f"Evaluating performance on {datasets[i]}...\n")
   evaluate(scores, test)
