import glob, os, subprocess, sys

sys.stderr.write(f"Setting up...\n")

import numpy as np
import pandas as pd
from functools import partial
from libraries.magma import Magma
from scipy.optimize import basinhopping
from multiprocessing import Pool, Manager
from sklearn.metrics import roc_curve, auc
from libraries import mlayerlinkpred as mlp

def fit_ensemble(df, methods, x):
   df["ensemble"] = 0
   for i in range(len(methods)):
      df["ensemble"] += (x[i] * df[methods[i]])
   fpr, tpr, thresholds = roc_curve(df["target"], df["ensemble"])
   return 1 - auc(fpr, tpr)

def _magma_apply_rules(m, antecedent):
   m.apply_rules(start = antecedent, end = antecedent)

def run_magma():
   m = Magma("data/pardus/pardus", 15, 4, directed = True, memory = 192)
   m.run()
   pool = Pool(processes = 32)
   _ = pool.map(partial(_magma_apply_rules, m), list(set(m.rules["antecedent"])))
   pool.close()
   pool.join()
   dfs = []
   for txt in glob.glob("data/pardus/*magmascore*"):
      dfs.append(pd.read_csv(txt, sep = "\t"))
      os.remove(txt)
   df = pd.concat(dfs)
   df.loc[df["src"] < 0, "src"] = df.loc[df["src"] < 0, "src"].map(lambda x: f"new{x*-1}")
   df.loc[df["trg"] < 0, "trg"] = df.loc[df["trg"] < 0, "trg"].map(lambda x: f"new{x*-1}")
   df = df.groupby(by = ["src", "trg", "layer"]).sum().reset_index()
   df["src"] = df["src"].astype(str)
   df["trg"] = df["trg"].astype(str)
   df["layer"] = df["layer"].astype(int)
   os.remove("data/pardus/pardus_15_4.mossout")
   os.remove("data/pardus/pardus_15_4.rules")
   return df[["src", "trg", "layer", "conf"]].rename(columns = {"conf": "old"})

def _moss_apply_rules(rules, peds, i):
   pd.DataFrame(rules.iloc[i]).T.to_csv(f"data/moss6_{i}.rul", sep = "\t", index = False)
   bash_command = f"java -Xmx6g -cp libraries/moss_new.jar moss.Predictor -fnel -D data/moss6.sub data/moss6_{i}.rul data/pardus/pardus.gspan data/moss6_{i}.ped"
   process = subprocess.Popen(bash_command.split(), stdout = subprocess.PIPE, stderr = subprocess.PIPE)
   output, error = process.communicate()
   ped = pd.read_csv(f"data/moss6_{i}.ped", sep = "\t")
   ped = ped.drop(["graph", "wgt", "lift1", "lift2"], axis = 1).groupby(by = ["src", "dst", "edge", "node"], dropna = False).sum().reset_index()
   ped = ped.rename(columns = {"dst": "trg", "edge": "layer", "conf": "new"})
   ped.loc[ped["src"] == 0, "src"] = ped.loc[ped["src"] == 0, "node"].map(lambda x : f"new{x:.0f}")
   ped.loc[ped["trg"] == 0, "trg"] = ped.loc[ped["trg"] == 0, "node"].map(lambda x : f"new{x:.0f}")
   ped["src"] = ped["src"].astype(str)
   ped["trg"] = ped["trg"].astype(str)
   ped["layer"] = ped["layer"].astype(int)
   ped = ped.merge(test, on = ["src", "trg", "layer"])
   if ped.shape[0] > 0:
      peds.append(ped[["src", "trg", "layer", "new"]])
   os.remove(f"data/moss6_{i}.rul")
   os.remove(f"data/moss6_{i}.ped")

def run_moss():
   bash_command = "java -Xmx192g -cp libraries/moss_new.jar moss.Miner -inel -onel -x -m2 -n4 -s6 -k3 -D -Zdata/moss6.rul -w0 -T data/pardus/pardus.gspan data/moss6.sub"
   process = subprocess.Popen(bash_command.split(), stdout = subprocess.PIPE, stderr = subprocess.PIPE)
   output, error = process.communicate()
   rules = pd.read_csv("data/moss6.rul", sep = "\t", dtype = {"node": str}, keep_default_na = False).astype(str)
   manager = Manager()
   peds = manager.list()
   pool = Pool(processes = 32)
   _ = pool.map(partial(_moss_apply_rules, rules, peds), range(rules.shape[0]))
   pool.close()
   pool.join()
   peds = pd.concat(list(peds))
   peds = peds.groupby(by = ["src", "trg", "layer"]).sum().reset_index()
   peds["src"] = peds["src"].astype(str)
   peds["trg"] = peds["trg"].astype(str)
   peds["layer"] = peds["layer"].astype(int)
   return peds

def ensemble(scores):
   methods = ["debacco", "hristova", "jalili", "old", "new", "pujari", "sharma", "mell"]
   for lp in methods:
      scores[lp] = (scores[lp] - scores[lp].mean()) / scores[lp].std()
   scores["ensemble_base"] = scores[methods].mean(axis = 1)
   df = pd.concat([scores[scores["target"] == 0].sample(2500), scores[scores["target"] == 1].sample(2500)])
   x0 = [1.0] * len(methods)
   res = basinhopping(partial(fit_ensemble, df, methods), x0)
   scores["ensemble_optim"] = 0
   for i in range(len(methods)):
      scores["ensemble_optim"] += (res.x[i] * scores[methods[i]])

def evaluate(test, fign, tabn, coln):
   evaluation = []
   for lp in ("debacco", "hristova", "jalili", "pujari", "sharma", "mell", "old", "new", "ensemble_base", "ensemble_optim"):
      test[lp] = test[lp].fillna(test[lp].min())
      fpr, tpr, thresholds = roc_curve(test["target"], test[lp])
      evaluation.append((lp, auc(fpr, tpr), fpr, tpr))
   evaluation = pd.DataFrame(data = evaluation, columns = ("method", "auc", "fpr", "tpr"))
   evaluation = evaluation.set_index("method")
   btc = evaluation.loc[["debacco", "hristova", "jalili", "pujari", "sharma"]].sort_values(by = "auc").index[-1]
   np.savetxt(f"fig_{fign}_red.csv", np.vstack((evaluation.loc["old", "fpr"], evaluation.loc["old", "tpr"])).T, delimiter = "\t", fmt = "%.8f")
   np.savetxt(f"fig_{fign}_green.csv", np.vstack((evaluation.loc["new", "fpr"], evaluation.loc["new", "tpr"])).T, delimiter = "\t", fmt = "%.8f")
   np.savetxt(f"fig_{fign}_blue.csv", np.vstack((evaluation.loc["mell", "fpr"], evaluation.loc["mell", "tpr"])).T, delimiter = "\t", fmt = "%.8f")
   np.savetxt(f"fig_{fign}_purple.csv", np.vstack((evaluation.loc[btc, "fpr"], evaluation.loc[btc, "tpr"])).T, delimiter = "\t", fmt = "%.8f")
   evaluation["auc"].reset_index().to_csv(f"tab{tabn}_col{coln}.csv", index = False, sep = "\t")

test = pd.read_csv("data/pardus/pardus_testset.csv", sep = "\t")
test_oldold = test[(test["src"] != "new1") & (test["src"] != "new2") & (test["trg"] != "new1") & (test["trg"] != "new2")].astype(int)
test_oldnew = test[~((test["src"] != "new1") & (test["src"] != "new2") & (test["trg"] != "new1") & (test["trg"] != "new2"))]

sys.stderr.write(f"Running Sharma MLP...\n")
test_oldold = test_oldold.merge(mlp.sharma("data/pardus/pardus", directed = True), on = ["src", "trg", "layer"], how = "left").fillna(0)
sys.stderr.write(f"Running Pujari MLP...\n")
test_oldold = test_oldold.merge(mlp.pujari("data/pardus/pardus", directed = True), on = ["src", "trg", "layer"], how = "left").fillna(0)
sys.stderr.write(f"Running Jalili MLP...\n")
test_oldold = test_oldold.merge(mlp.jalili("data/pardus/pardus", directed = True), on = ["src", "trg", "layer"], how = "left").fillna(0)
sys.stderr.write(f"Running Hristova MLP...\n")
test_oldold = test_oldold.merge(mlp.hristova("data/pardus/pardus", directed = True), on = ["src", "trg", "layer"], how = "left").fillna(0)
sys.stderr.write(f"Running De Bacco MLP...\n")
test_oldold = test_oldold.merge(mlp.debacco("data/pardus/pardus", directed = True), on = ["src", "trg", "layer"], how = "left").fillna(0)
sys.stderr.write(f"Running MELL MLP...\n")
test_oldold = test_oldold.merge(mlp.mell("data/pardus/pardus", directed = True), on = ["src", "trg", "layer"], how = "left").fillna(0)

test_oldold["src"] = test_oldold["src"].astype(str)
test_oldold["trg"] = test_oldold["trg"].astype(str)

test = pd.concat([test_oldold, test_oldnew]).fillna(0)

sys.stderr.write(f"Running old code MLP...\n")
test = test.merge(run_magma(), on = ["src", "trg", "layer"], how = "left").fillna(0)
sys.stderr.write(f"Running new code MLP...\n")
test = test.merge(run_moss(), on = ["src", "trg", "layer"], how = "left").fillna(0)

sys.stderr.write(f"Running ensemble MLP...\n")
ensemble(test)

sys.stderr.write(f"Evaluating performance...\n")
test_oldold = pd.read_csv("data/pardus/pardus_testset_oldold.csv", sep = "\t", dtype = {"src": str, "trg": str}).drop("target", axis = 1).merge(test, on = ["src", "trg", "layer"])
test_oldnew = pd.read_csv("data/pardus/pardus_testset_oldnew.csv", sep = "\t").drop("target", axis = 1).merge(test, on = ["src", "trg", "layer"])

evaluate(test, "12d", 2, 4)
evaluate(test_oldold, "13a", 3, 1)
evaluate(test_oldnew, "13b", 3, 2)

