import os, glob, subprocess, sys
import numpy as np
import pandas as pd
import networkx as nx
from functools import partial
from libraries import node2vec
from gensim.models import Word2Vec
from scipy.optimize import basinhopping
from multiprocessing import Pool, Manager
from sklearn.metrics import roc_curve, auc
from libraries import mlayerlinkpred as mlp
from sklearn.linear_model import LogisticRegressionCV

def fit_ensemble(df, methods, x):
   df["ensemble"] = 0
   for i in range(len(methods)):
      df["ensemble"] += (x[i] * df[methods[i]])
   fpr, tpr, thresholds = roc_curve(df["target"], df["ensemble"])
   return 1 - auc(fpr, tpr)

def _moss_wrapper(x):
   s = {
      "aarhus": 40,
      "phys": 2,
      "celegans": 96
   }
   d = "-D" if x[0] != "aarhus" else ""
   bash_command = f"java -Xmx16g -cp libraries/moss_new.jar moss.Miner -inel -onel -x -m2 -n4 -s{s[x[0]]} -k3 {d} -Zdata/moss{x[1]}.rul -w0 -T data/{x[0]}/{x[0]}_sl{x[1]}.gspan data/moss{x[1]}.sub"
   process = subprocess.Popen(bash_command.split(), stdout = subprocess.PIPE, stderr = subprocess.PIPE)
   output, error = process.communicate()
   bash_command = f"java -Xmx16g -cp libraries/moss_new.jar moss.Predictor -fnel {d} data/moss{x[1]}.sub data/moss{x[1]}.rul data/{x[0]}/{x[0]}_sl{x[1]}.gspan data/moss{x[1]}.ped"
   process = subprocess.Popen(bash_command.split(), stdout = subprocess.PIPE, stderr = subprocess.PIPE)
   output, error = process.communicate()

def moss_wrapper(network):
   pool = Pool(processes = nproc[network])
   _ = pool.map(_moss_wrapper, [(network, fold) for fold in range(10)])
   pool.close()
   pool.join()
   dfs = []
   for txt in glob.glob("data/moss*.ped"):
      netid = txt.split("moss")[1].split('.')[0].split('_')[0]
      df = pd.read_csv(txt, sep = "\t")
      df = df.rename(columns = {"dst": "trg", "edge": "layer", "wgt": "count", "lift1": "lift"})
      df.loc[df["src"] == 0, "src"] = df.loc[df["src"] == 0, "node"].map(lambda x : f"new{x:.0f}")
      df.loc[df["trg"] == 0, "trg"] = df.loc[df["trg"] == 0, "node"].map(lambda x : f"new{x:.0f}")
      df["fold"] = int(netid)
      dfs.append(df.drop(["graph", "node", "lift2", "count", "lift"], axis = 1))
   for f in glob.glob(f"data/moss*"):
      os.remove(f)
   df = pd.concat(dfs)
   df = df.groupby(by = ["src", "trg", "fold"]).sum().reset_index().rename(columns = {"conf": "new"})
   df["src"] = df["src"].astype(str)
   df["trg"] = df["trg"].astype(str)
   return df

def _magma_wrapper(x):
   s = {
      "aarhus": 10,
      "phys": 20,
      "celegans": 100
   }
   mlp.magma(f"data/{x[0]}/{x[0]}_sl{x[1]}", s[x[0]], 4, directed = x[0] != "aarhus")

def magma_wrapper(network):
   pool = Pool(processes = nproc[network])
   _ = pool.map(_magma_wrapper, [[network, fold] for fold in range(10)])
   pool.close()
   pool.join()
   dfs = []
   for txt in glob.glob(f"data/{network}/*magmascore*"):
      netid = txt.split("_")[1][-1]
      df = pd.read_csv(txt, sep = "\t")
      df["fold"] = int(netid)
      df.loc[df["src"] < 0, "src"] = df.loc[df["src"] < 0, "src"].map(lambda x: f"new{x*-1}")
      df.loc[df["trg"] < 0, "trg"] = df.loc[df["trg"] < 0, "trg"].map(lambda x: f"new{x*-1}")
      dfs.append(df)
      os.remove(txt)
   for f in glob.glob(f"data/{network}/*.rules"):
      os.remove(f)
   for f in glob.glob(f"data/{network}/*.mossout"):
      os.remove(f)
   df = pd.concat(dfs)
   df = df.groupby(by = ["src", "trg", "fold"]).sum().reset_index()
   df["src"] = df["src"].astype(str)
   df["trg"] = df["trg"].astype(str)
   return df.rename(columns = {"magma": "old"})

def n2v_wrapper(G, x):
   directed = x[0] != "aarhus"
   nodes = set(G.nodes)
   edges = set(G.edges)
   for edge in G.edges():
      G[edge[0]][edge[1]]["weight"] = 1
   G = node2vec.Graph(G, directed, 1, 1)
   G.preprocess_transition_probs()
   walks = G.simulate_walks(10, 80)
   walks = [list(map(str, walk)) for walk in walks]
   model = Word2Vec(sentences = walks, vector_size = 128, window = 10, min_count = 0, sg = 1, workers = 8, epochs = 1)
   if directed:
      scores = [(src, trg, 1 if (src, trg) in edges else 0) for src in nodes for trg in nodes if src != trg]
   else:
      scores = [(src, trg, 1 if (src, trg) in edges or (trg, src) in edges else 0) for src in nodes for trg in nodes if int(src) < int(trg)]
   scores = pd.DataFrame(data = scores, columns = ("src", "trg", "target"))
   scores[["src", "trg"]] = scores[["src", "trg"]].astype(str)
   scores["lhs"] = scores.apply(lambda x: np.abs(model.wv[model.wv.key_to_index[x["src"]]] - model.wv[model.wv.key_to_index[x["trg"]]]), axis = 1)
   training = pd.concat([scores[scores["target"] == 1], scores[scores["target"] == 0].sample(len(edges))])
   lr_clf = LogisticRegressionCV(Cs = 10, cv = 10, scoring = "roc_auc", max_iter = 2000)
   lr_clf.fit(np.concatenate(training["lhs"].values).reshape(training.shape[0], 128), training["target"])
   scores["n2v"] = lr_clf.predict_proba(np.concatenate(scores["lhs"].values).reshape(scores.shape[0], 128))[:,1]
   return scores[scores["target"] == 0].drop(["target", "lhs"], axis = 1)

def traditional_wrapper(x):
   directed = x[0] != "aarhus"
   G = nx.read_edgelist(f"data/{x[0]}/{x[0]}_sl{x[1]}", nodetype = int, create_using = nx.DiGraph() if directed else nx.Graph(), data = [("label", int),])
   fold_df = pd.DataFrame(data = list(nx.resource_allocation_index(nx.Graph(G))), columns = ("src", "trg", "ra"))
   fold_df = fold_df.merge(pd.DataFrame(data = list(nx.adamic_adar_index(nx.Graph(G))), columns = ("src", "trg", "aa")), on = ["src", "trg"], how = "outer")
   fold_df = fold_df.merge(pd.DataFrame(data = list(nx.preferential_attachment(nx.Graph(G))), columns = ("src", "trg", "pa")), on = ["src", "trg"], how = "outer")
   fold_df = fold_df.merge(pd.DataFrame(data = list(nx.jaccard_coefficient(nx.Graph(G))), columns = ("src", "trg", "ja")), on = ["src", "trg"], how = "outer")
   fold_df[["src", "trg"]] = fold_df[["src", "trg"]].astype(str)
   fold_df = fold_df.merge(n2v_wrapper(G, x), on = ["src", "trg"], how = "outer")
   fold_df["fold"] = x[1]
   x[2].append(fold_df.fillna(0))

def ensemble(scores, network):
   methods = ["ra", "aa", "ja", "pa", "n2v", "old", "new"]
   for lp in methods:
      scores[lp] = (scores[lp] - scores[lp].mean()) / scores[lp].std()
   scores["ensemble_base"] = scores[methods].mean(axis = 1)
   df = test_set(network).merge(scores, on = ["src", "trg", "fold"], how = "left")
   for lp in methods:
      df[lp] = df[lp].fillna(df[lp].min())
   x0 = [1.0] * len(methods)
   res = basinhopping(partial(fit_ensemble, df, methods), x0)
   scores["ensemble_optim"] = 0
   for i in range(len(methods)):
      scores["ensemble_optim"] += (res.x[i] * scores[methods[i]])

def run_mlps(network):
   sys.stderr.write(f"Running all traditional MLPs on {network} in parallel...\n")
   manager = Manager()
   results = manager.list()
   pool = Pool(processes = nproc[network])
   _ = pool.map(traditional_wrapper, [[network, fold, results] for fold in range(10)])
   pool.close()
   pool.join()
   results = pd.concat(list(results))
   if network == "phys":
      results["pa"] *= -1
   elif network == "celegans":
      results[["ja", "ra", "aa", "pa"]] *= -1
   sys.stderr.write(f"Running old code on {network}...\n")
   results = results.merge(magma_wrapper(network), on = ["src", "trg", "fold"], how = "outer").fillna(0)
   sys.stderr.write(f"Running new code on {network}...\n")
   results = results.merge(moss_wrapper(network), on = ["src", "trg", "fold"], how = "outer").fillna(0)
   sys.stderr.write(f"Running ensemble on {network}...\n")
   ensemble(results, network)
   return results

def test_set(network):
   tests = []
   directed = network != "aarhus"
   for fold in range(10):
      target = pd.read_csv(f"data/{network}/{network}_sl{fold}_test", sep = "\t", header = None, names = ("src", "trg", "layer", "fold"), dtype = {"src": str, "trg": str}).drop("layer", axis = 1)
      df2 = pd.read_csv(f"data/{network}/{network}_sl{fold}", sep = "\t", header = None, names = ("src", "trg", "layer"), dtype = {"src": str, "trg": str}).drop("layer", axis = 1)
      df2["fold"] = fold
      srcs = set(df2["src"])
      trgs = set(df2["trg"])
      if directed:
         edges = [(src, trg, fold) for src in srcs for trg in trgs if src != trg]
      else:
         edges = [(src, trg, fold) for src in srcs for trg in trgs if int(src) < int(trg)]
      test = pd.DataFrame(data = edges, columns = ("src", "trg", "fold"))
      test = pd.concat([test, target, df2])
      test = test.drop_duplicates(keep = False)
      test = test.sample(target.shape[0])
      test["target"] = 0
      target["target"] = 1
      tests.append(pd.concat([test, target]))
   return pd.concat(tests)

def evaluate(scores, test):
   test = test.merge(scores, on = ["src", "trg", "fold"], how = "left")
   evaluation = []
   for lp in ("ra", "aa", "ja", "pa", "n2v", "old", "new", "ensemble_base", "ensemble_optim"):
      test[lp] = test[lp].fillna(test[lp].min())
      fpr, tpr, thresholds = roc_curve(test["target"], test[lp])
      evaluation.append((lp, auc(fpr, tpr), fpr, tpr))
   evaluation = pd.DataFrame(data = evaluation, columns = ("method", "auc", "fpr", "tpr"))
   evaluation = evaluation.set_index("method")
   btc = evaluation.loc[["ra", "aa", "ja", "pa"]].sort_values(by = "auc").index[-1]
   np.savetxt(f"fig_14{figlabel[i]}_red.csv", np.vstack((evaluation.loc["old", "fpr"], evaluation.loc["old", "tpr"])).T, delimiter = "\t", fmt = "%.8f")
   np.savetxt(f"fig_14{figlabel[i]}_green.csv", np.vstack((evaluation.loc["new", "fpr"], evaluation.loc["new", "tpr"])).T, delimiter = "\t", fmt = "%.8f")
   np.savetxt(f"fig_14{figlabel[i]}_blue.csv", np.vstack((evaluation.loc["n2v", "fpr"], evaluation.loc["n2v", "tpr"])).T, delimiter = "\t", fmt = "%.8f")
   np.savetxt(f"fig_14{figlabel[i]}_purple.csv", np.vstack((evaluation.loc[btc, "fpr"], evaluation.loc[btc, "tpr"])).T, delimiter = "\t", fmt = "%.8f")
   evaluation["auc"].reset_index().to_csv(f"tab4_col{i + 1}.csv", index = False, sep = "\t")

nproc = {
  "aarhus": 10,
  "phys": 10,
  "celegans": 2
}

datasets = ["aarhus", "phys", "celegans"]
figlabel = "abc"

for i in range(len(datasets)):
   scores = run_mlps(datasets[i])
   sys.stderr.write(f"Building test set of {datasets[i]}...\n")
   test = test_set(datasets[i])
   sys.stderr.write(f"Evaluating performance on {datasets[i]}...\n")
   evaluate(scores, test)
