import glob, os, subprocess, sys

sys.stderr.write(f"Setting up...\n")

import numpy as np
import pandas as pd
import networkx as nx
from functools import partial
from libraries import node2vec
from libraries.magma import Magma
from gensim.models import Word2Vec
from scipy.optimize import basinhopping
from multiprocessing import Pool, Manager
from sklearn.metrics import roc_curve, auc
from sklearn.linear_model import LogisticRegressionCV

def fit_ensemble(df, methods, x):
   df["ensemble"] = 0
   for i in range(len(methods)):
      df["ensemble"] += (x[i] * df[methods[i]])
   fpr, tpr, thresholds = roc_curve(df["target"], df["ensemble"])
   return 1 - auc(fpr, tpr)

def _magma_apply_rules(m, antecedent):
   m.apply_rules(start = antecedent, end = antecedent)

def run_magma():
   m = Magma("data/pardus/pardus_sl", 20, 4, directed = True, memory = 192)
   m.run()
   pool = Pool(processes = 32)
   _ = pool.map(partial(_magma_apply_rules, m), list(set(m.rules["antecedent"])))
   pool.close()
   pool.join()
   dfs = []
   for txt in glob.glob("data/pardus/*magmascore*"):
      dfs.append(pd.read_csv(txt, sep = "\t"))
      os.remove(txt)
   df = pd.concat(dfs)
   df.loc[df["src"] < 0, "src"] = df.loc[df["src"] < 0, "src"].map(lambda x: f"new{x*-1}")
   df.loc[df["trg"] < 0, "trg"] = df.loc[df["trg"] < 0, "trg"].map(lambda x: f"new{x*-1}")
   df = df.groupby(by = ["src", "trg", "layer"]).sum().reset_index()
   df["src"] = df["src"].astype(str)
   df["trg"] = df["trg"].astype(str)
   df["layer"] = df["layer"].astype(int)
   os.remove("data/pardus/pardus_sl_20_4.mossout")
   os.remove("data/pardus/pardus_sl_20_4.rules")
   return df[["src", "trg", "conf"]].rename(columns = {"conf": "old"})

def _moss_apply_rules(rules, peds, i):
   pd.DataFrame(rules.iloc[i]).T.to_csv(f"data/moss20_{i}.rul", sep = "\t", index = False)
   bash_command = f"java -Xmx6g -cp libraries/moss_new.jar moss.Predictor -fnel -D data/moss20.sub data/moss20_{i}.rul data/pardus/pardus_sl.gspan data/moss20_{i}.ped"
   process = subprocess.Popen(bash_command.split(), stdout = subprocess.PIPE, stderr = subprocess.PIPE)
   output, error = process.communicate()
   ped = pd.read_csv(f"data/moss20_{i}.ped", sep = "\t")
   ped = ped.drop(["graph", "wgt", "lift1", "lift2"], axis = 1).groupby(by = ["src", "dst", "edge", "node"], dropna = False).sum().reset_index()
   ped = ped.rename(columns = {"dst": "trg", "edge": "layer", "conf": "new"}).drop("layer", axis = 1)
   ped.loc[ped["src"] == 0, "src"] = ped.loc[ped["src"] == 0, "node"].map(lambda x : f"new{x:.0f}")
   ped.loc[ped["trg"] == 0, "trg"] = ped.loc[ped["trg"] == 0, "node"].map(lambda x : f"new{x:.0f}")
   ped["src"] = ped["src"].astype(str)
   ped["trg"] = ped["trg"].astype(str)
   ped = ped.merge(test, on = ["src", "trg"])
   if ped.shape[0] > 0:
      peds.append(ped[["src", "trg", "new"]])
   os.remove(f"data/moss20_{i}.rul")
   os.remove(f"data/moss20_{i}.ped")

def run_moss():
   bash_command = "java -Xmx192g -cp libraries/moss_new.jar moss.Miner -inel -onel -x -m2 -n4 -s20 -k3 -D -Zdata/moss20.rul -w0 -T data/pardus/pardus_sl.gspan data/moss20.sub"
   process = subprocess.Popen(bash_command.split(), stdout = subprocess.PIPE, stderr = subprocess.PIPE)
   output, error = process.communicate()
   rules = pd.read_csv("data/moss20.rul", sep = "\t", dtype = {"node": str}, keep_default_na = False).astype(str)
   manager = Manager()
   peds = manager.list()
   pool = Pool(processes = 32)
   _ = pool.map(partial(_moss_apply_rules, rules, peds), range(rules.shape[0]))
   pool.close()
   pool.join()
   peds = pd.concat(list(peds))
   peds = peds.groupby(by = ["src", "trg"]).sum().reset_index()
   return peds

def n2v_wrapper(G, test):
   def assign_lhs(x, model):
      if x["src"] in model.key_to_index and x["trg"] in model.key_to_index:
         return np.abs(model[model.key_to_index[x["src"]]] - model[model.key_to_index[x["trg"]]])
      else:
         return np.zeros(128)
   nodes = set(G.nodes)
   edges = set(G.edges)
   for edge in G.edges():
      G[edge[0]][edge[1]]["weight"] = 1
   G = node2vec.Graph(G, True, 1, 1)
   G.preprocess_transition_probs()
   walks = G.simulate_walks(10, 80)
   walks = [list(map(str, walk)) for walk in walks]
   model = Word2Vec(sentences = walks, vector_size = 128, window = 10, min_count = 0, sg = 1, workers = 8, epochs = 1)
   scores = [(src, trg, 1 if (src, trg) in edges else 0) for src in nodes for trg in nodes if src != trg]
   scores = pd.DataFrame(data = scores, columns = ("src", "trg", "target"))
   scores = pd.concat([scores[scores["target"] == 1], scores[scores["target"] == 0].sample(scores[scores["target"] == 1].shape[0])])
   scores["lhs"] = scores.apply(lambda x: np.abs(model.wv[model.wv.key_to_index[x["src"]]] - model.wv[model.wv.key_to_index[x["trg"]]]), axis = 1)
   lr_clf = LogisticRegressionCV(Cs = 10, cv = 10, scoring = "roc_auc", max_iter = 2000)
   lr_clf.fit(np.concatenate(scores["lhs"].values).reshape(scores.shape[0], 128), scores["target"])
   test["lhs"] = test.apply(assign_lhs, axis = 1, args = (model.wv,))
   test["n2v"] = lr_clf.predict_proba(np.concatenate(test["lhs"].values).reshape(test.shape[0], 128))[:,1]
   test.loc[test["src"].isin(("new1", "new2")) | test["trg"].isin(("new1", "new2")), "n2v"] = 0
   test = test.drop("lhs", axis = 1)

def ensemble(scores):
   methods = ["aa", "ra", "ja", "old", "new", "pa", "n2v"]
   for lp in methods:
      scores[lp] = (scores[lp] - scores[lp].mean()) / scores[lp].std()
   scores["ensemble_base"] = scores[methods].mean(axis = 1)
   df = pd.concat([scores[scores["target"] == 0].sample(1500), scores[scores["target"] == 1].sample(1500)])
   x0 = [1.0] * len(methods)
   res = basinhopping(partial(fit_ensemble, df, methods), x0)
   scores["ensemble_optim"] = 0
   for i in range(len(methods)):
      scores["ensemble_optim"] += (res.x[i] * scores[methods[i]])

test = pd.read_csv("data/pardus/pardus_sl_testset.csv", sep = "\t")
G = nx.read_edgelist(f"data/pardus/pardus_sl", nodetype = str, create_using = nx.DiGraph(), data = [("label", int),])

sys.stderr.write(f"Running Preferential Attachment LP...\n")
test = test.merge(pd.DataFrame(data = list(nx.preferential_attachment(nx.Graph(G))), columns = ("src", "trg", "pa")), on = ["src", "trg"], how = "left").fillna(0)
sys.stderr.write(f"Running Resource Allocation LP...\n")
test = test.merge(pd.DataFrame(data = list(nx.resource_allocation_index(nx.Graph(G))), columns = ("src", "trg", "ra")), on = ["src", "trg"], how = "left").fillna(0)
sys.stderr.write(f"Running Adamic Adar LP...\n")
test = test.merge(pd.DataFrame(data = list(nx.adamic_adar_index(nx.Graph(G))), columns = ("src", "trg", "aa")), on = ["src", "trg"], how = "left").fillna(0)
sys.stderr.write(f"Running Jaccard LP...\n")
test = test.merge(pd.DataFrame(data = list(nx.jaccard_coefficient(nx.Graph(G))), columns = ("src", "trg", "ja")), on = ["src", "trg"], how = "left").fillna(0)

sys.stderr.write(f"Running Node2Vec LP...\n")
n2v_wrapper(G, test)

sys.stderr.write(f"Running old code LP...\n")
test = test.merge(run_magma(), on = ["src", "trg"], how = "left").fillna(0)
sys.stderr.write(f"Running new code LP...\n")
test = test.merge(run_moss(), on = ["src", "trg"], how = "left").fillna(0)

sys.stderr.write(f"Running Ensemble LP...\n")
ensemble(test)

sys.stderr.write(f"Evaluating...\n")
evaluation = []
for lp in ("aa", "ra", "ja", "old", "new", "pa", "n2v", "ensemble_base", "ensemble_optim"):
   test[lp] = test[lp].fillna(test[lp].min())
   fpr, tpr, thresholds = roc_curve(test["target"], test[lp])
   evaluation.append((lp, auc(fpr, tpr), fpr, tpr))

evaluation = pd.DataFrame(data = evaluation, columns = ("method", "auc", "fpr", "tpr"))
evaluation = evaluation.set_index("method")
btc = evaluation.loc[["aa", "ra", "ja", "pa"]].sort_values(by = "auc").index[-1]
np.savetxt(f"fig_14d_red.csv", np.vstack((evaluation.loc["old", "fpr"], evaluation.loc["old", "tpr"])).T, delimiter = "\t", fmt = "%.8f")
np.savetxt(f"fig_14d_green.csv", np.vstack((evaluation.loc["new", "fpr"], evaluation.loc["new", "tpr"])).T, delimiter = "\t", fmt = "%.8f")
np.savetxt(f"fig_14d_blue.csv", np.vstack((evaluation.loc["n2v", "fpr"], evaluation.loc["n2v", "tpr"])).T, delimiter = "\t", fmt = "%.8f")
np.savetxt(f"fig_14d_purple.csv", np.vstack((evaluation.loc[btc, "fpr"], evaluation.loc[btc, "tpr"])).T, delimiter = "\t", fmt = "%.8f")
evaluation["auc"].reset_index().to_csv(f"tab4_col4.csv", index = False, sep = "\t")

