import sys, random
import numpy as np
import pandas as pd
import networkx as nx
from simanneal import Annealer

class Frustration(Annealer):
   def __init__(self, G):
      self.G = G
      self.frustrated_edges = []
      super(Frustration, self).__init__(np.ones(len(G.nodes)))
      self.updates = 0
   
   def energy(self):
      self.frustrated_edges = []
      result = 0
      for edge in self.G.edges(data = True):
         if not ((edge[2]["label"] == 3 and self.state[edge[0]] == self.state[edge[1]]) or (edge[2]["label"] in (1, 2) and self.state[edge[0]] != self.state[edge[1]])):
            result += 1
            self.frustrated_edges.append(edge)
      if result == 0:
         self.user_exit = True
         return 0
      else:
         return result / len(self.G.edges)

   def move(self):
      self.state[random.choice(self.frustrated_edges)[random.randint(0, 1)]] *= -1

def patterns(infile):
   pattern = nx.MultiDiGraph()
   attrs = {}
   for line in infile:
      fields = line.strip().split()
      if len(fields) == 0:
         pattern.graph.update(attrs)
         yield pattern
         pattern = nx.DiGraph()
         attrs = {}
      elif fields[0] == 'n':
         pattern.add_node(int(fields[1]) - 1, label = int(fields[2]))
      elif fields[0] == 'e':
         pattern.add_edge(int(fields[1]) - 1, int(fields[2]) - 1, label = int(fields[3]))
      elif fields[0] == 'g':
         attrs["id"] = int(fields[1])
      elif fields[0] == 's':
         attrs["support"] = int(fields[3])

def frustration(G):
   frustr_anneal = Frustration(G)
   if frustr_anneal.energy() == 0:
      return 0
   else:
      p, score = frustr_anneal.anneal()
      return score

rulefile = "moss6"

i = 1
frustr = []
with open(f"data/{rulefile}.sub", 'r') as f:
   for pattern in patterns(f):
      frustr.append((i, frustration(pattern)))
      sys.stderr.write(f"\rCalculating frustration pattern {i}")
      sys.stderr.flush()
      i += 1

sys.stderr.write("\nCalculating summary statistics for rule frustrations\n")
frustr = pd.DataFrame(data = frustr, columns = ("pattern", "frustration"))
rules = pd.read_csv(f"data/{rulefile}.rul", sep = "\t")

rules = rules.merge(frustr, left_on = "part", right_on = "pattern")
rules = rules.merge(frustr, left_on = "full", right_on = "pattern", suffixes = ("_part", "_full"))
rules["frustration_delta"] = rules["frustration_full"] - rules["frustration_part"]

lfr_conf = rules[rules["frustration_delta"] <= 0]["conf"].round(1).value_counts().reset_index().sort_values(by = "index")
hfr_conf = rules[rules["frustration_delta"] > 0]["conf"].round(1).value_counts().reset_index().sort_values(by = "index")

lfr_conf["conf"] /= lfr_conf["conf"].sum()
hfr_conf["conf"] /= hfr_conf["conf"].sum()

lfr_conf.to_csv("fig_17a_green.csv", index = False, sep = "\t")
hfr_conf.to_csv("fig_17a_red.csv", index = False, sep = "\t")

lfr_supp = rules[rules["frustration_delta"] <= 0].groupby(by = "supp").size().reset_index().rename(columns = {0: "count"})
hfr_supp = rules[rules["frustration_delta"] > 0].groupby(by = "supp").size().reset_index().rename(columns = {0: "count"})

lfr_supp = lfr_supp.sort_values(by = "supp", ascending = False)
lfr_supp["cdf"] = lfr_supp["count"].cumsum() / lfr_supp["count"].sum()
lfr_supp.to_csv("fig_17b_green.csv", index = False, sep = "\t")

hfr_supp = hfr_supp.sort_values(by = "supp", ascending = False)
hfr_supp["cdf"] = hfr_supp["count"].cumsum() / hfr_supp["count"].sum()
hfr_supp.to_csv("fig_17b_red.csv", index = False, sep = "\t")
