import subprocess, os, sys
import numpy as np
import pandas as pd
import graph_tool as gt
from collections import defaultdict
from graph_tool.topology import subgraph_isomorphism as isom

class Magma():
   def __init__(self, infile, min_sup, max_size, min_conf = 0, directed = False, memory = 6):
      self.netname = infile
      self.min_sup = min_sup
      self.max_size = max_size
      self.directed = directed
      self.network = self.read_network(infile)
      self.patterns_by_id = {}
      self.patterns_by_size = defaultdict(list)
      self.memory = memory
      self.min_conf = min_conf

   def read_network(self, filename):
      network = gt.Graph(directed = self.directed)
      node_labels = []
      edge_labels = []
      with open("%s_nodeattrs" % filename, 'r') as f:
         for line in f:
            fields = line.strip().split("\t")
            node_labels.append(int(fields[1]))
      with open(filename, 'r') as f:
         for line in f:
            fields = [int(x) for x in line.strip().split("\t")]
            if self.directed:
               network.add_edge(fields[0], fields[1])
            else:
               network.add_edge(min(fields[0], fields[1]), max(fields[0], fields[1]))
            edge_labels.append(fields[2])
      if self.directed:
         isolated = (network.get_out_degrees(list(network.vertices())) + network.get_in_degrees(list(network.vertices()))) == 0
      else:
         isolated = network.get_out_degrees(list(network.vertices())) == 0
      nnz = (~isolated).nonzero()[0]
      self.nodemap = {i: nnz[i] for i in range(nnz.shape[0])}
      for node_label in set(node_labels):
         self.nodemap[-node_label] = -node_label
      network.remove_vertex(np.where(isolated)[0])
      network.vp.label = network.new_vertex_property("int", vals = node_labels)
      network.ep.layer = network.new_edge_property("int", vals = edge_labels)
      return network

   def run(self):
      self.make_moss_input()
      self.run_moss()
      self.read_patterns()
      self.build_rules()

   def make_moss_input(self):
      with open("%s_%s_%s.gspan" % (self.netname, self.min_sup, self.max_size), 'w') as f:
         for v in self.network.vertices():
            f.write("v %s %s\n" % (int(v) + 1, self.network.vp.label[v]))
         for e in self.network.edges():
            f.write("e %s %s %s\n" % (int(e.source()) + 1, int(e.target()) + 1, self.network.ep.layer[e]))
   
   def run_moss(self):
      infile = "%s_%s_%s.gspan" % (self.netname, self.min_sup, self.max_size)
      outfile = "%s_%s_%s.mossout" % (self.netname, self.min_sup, self.max_size)
      bash_command = "java -Xmx%sg -cp libraries/moss_old.jar moss.Miner -inel -onel -x -m2 -n%s -s%s -k3 -C -M1 %s %s" % (self.memory, self.max_size, self.min_sup, infile, outfile)
      process = subprocess.Popen(bash_command.split(), stdout = subprocess.PIPE, stderr = subprocess.PIPE)
      output, error = process.communicate()
      os.remove(infile)

   def read_patterns(self):
      with open("%s_%s_%s.mossout" % (self.netname, self.min_sup, self.max_size), 'r') as f:
         G = gt.Graph(directed = self.directed)
         node_labels = []
         edge_labels = []
         for line in f:
            if line.strip() == "":
               G.vp.label = G.new_vertex_property("int", vals = node_labels)
               G.ep.layer = G.new_edge_property("int", vals = edge_labels)
               if self.directed:
                  isolated = (G.get_out_degrees(list(G.vertices())) + G.get_in_degrees(list(G.vertices()))) == 0
               else:
                  isolated = G.get_out_degrees(list(G.vertices())) == 0
               G.remove_vertex(np.where(isolated)[0])
               self.patterns_by_size[nedges].append((G, supp, pid))
               self.patterns_by_id[pid] = G
               G = gt.Graph(directed = self.directed)
               node_labels = []
               edge_labels = []
            else:
               fields = line.strip().split()
               if line[0] == 'n':
                  node_labels.append(int(fields[-1]))
               elif line[0] == 'e':
                  if self.directed:
                     G.add_edge(int(fields[1]) - 1, int(fields[2]) - 1)
                  else:
                     G.add_edge(int(min(fields[1], fields[2])) - 1, int(max(fields[1], fields[2])) - 1)
                  edge_labels.append(int(fields[-1]))
               elif line[0] == 'g':
                  pid = int(fields[-1])
               elif line[0] == 's':
                  supp = int(fields[3])
                  nedges = int(fields[2])

   def find_missing_layer(self, g, sub, vmap):
      for v in sub.vertices():
         w = g.vertex(vmap[v])
         us = defaultdict(set)
         for x in v.out_edges():
            us[g.vertex(vmap[x.target()])].add(sub.ep.layer[x])
         for ew in w.out_edges():
            if not (ew.target() in us and g.ep.layer[ew] in us[ew.target()]):
               return g.ep.layer[ew]
         if self.directed:
            us = defaultdict(set)
            for x in v.in_edges():
               us[g.vertex(vmap[x.source()])].add(sub.ep.layer[x])
            for ew in w.in_edges():
               if not (ew.source() in us and g.ep.layer[ew] in us[ew.source()]):
                  return g.ep.layer[ew]

   def build_rules(self):
      self.layer_support = {p[0].ep.layer.a[0]: p[1] / self.network.num_vertices() for p in self.patterns_by_size[1]}
      rules = []
      for nedges in sorted(self.patterns_by_size.keys())[:-1]:
         #sys.stderr.write("# Edges: %s->%s (total tests: %s)\n" % (nedges, nedges + 1, len(self.patterns_by_size[nedges]) * len(self.patterns_by_size[nedges + 1])))
         #sys.stderr.flush()
         for a in self.patterns_by_size[nedges]:
            for c in self.patterns_by_size[nedges + 1]:
               gm = isom(a[0], c[0], max_n = 1, vertex_label = (a[0].vp.label, c[0].vp.label), edge_label = (a[0].ep.layer, c[0].ep.layer))
               if len(gm) > 0:
                  layer = self.find_missing_layer(c[0], a[0], gm[0])
                  confidence = c[1] / a[1]
                  if confidence >= self.min_conf:
                     lift = confidence / self.layer_support[layer]
                     rules.append((a[2], c[2], c[1], confidence, lift))
      self.rules = pd.DataFrame(data = rules, columns = ("antecedent", "consequent", "support", "confidence", "lift"))
      self.rules.sort_values(by = ["support", "confidence"], ascending = False).to_csv("%s_%s_%s.rules" % (self.netname, self.min_sup, self.max_size), sep = "\t", index = False)

   def find_edge(self, sub, g, sub_g_vmap, g_sub_vmap):
      for v in sub.vertices():
         w = g.vertex(sub_g_vmap[v])
         us = defaultdict(set)
         for x in v.out_edges():
            us[g.vertex(sub_g_vmap[x.target()])].add(sub.ep.layer[x])
         for ew in w.out_edges():
            if not (ew.target() in us and g.ep.layer[ew] in us[ew.target()]):
               if ew.target() in g_sub_vmap:
                  return (w, g_sub_vmap[ew.target()], g.ep.layer[ew])
               else:
                  return (w, -g.vp.label[ew.target()], g.ep.layer[ew])
         if self.directed:
            us = defaultdict(set)
            for x in v.in_edges():
               us[g.vertex(sub_g_vmap[x.source()])].add(sub.ep.layer[x])
            for ew in w.in_edges():
               if not (ew.source() in us and g.ep.layer[ew] in us[ew.source()]):
                  if ew.source() in g_sub_vmap:
                     return (g_sub_vmap[ew.source()], w, g.ep.layer[ew])
                  else:
                     return (-g.vp.label[ew.source()], w, g.ep.layer[ew])

   def apply_rules(self, start = -1, end = 1e6):
      edge_confs = defaultdict(float)
      edge_lifts = defaultdict(float)
      edge_counts = defaultdict(int)
      for a in set(self.rules["antecedent"]):
         if a >= start and a <= end:
            #sys.stderr.write("Antecedent #%s rules...\n" % a)
            for _, consequent_row in self.rules[self.rules["antecedent"] == a].iterrows():
               c = self.patterns_by_id[consequent_row["consequent"]]
               ac_mapper = isom(self.patterns_by_id[a], c, max_n = 1, vertex_label = (self.patterns_by_id[a].vp.label, c.vp.label), edge_label = (self.patterns_by_id[a].ep.layer, c.ep.layer))[0]
               ca_mapper = {ac_mapper[i]: i for i in range(ac_mapper.a.shape[0])}
               extra_edge = self.find_edge(self.patterns_by_id[a], c, ac_mapper, ca_mapper)
               vs = isom(self.patterns_by_id[a], self.network, vertex_label = (self.patterns_by_id[a].vp.label, self.network.vp.label), edge_label = (self.patterns_by_id[a].ep.layer, self.network.ep.layer))
               for vmap in vs:
                  try:
                     edge = (vmap[extra_edge[0]], vmap[extra_edge[1]], extra_edge[2])
                  except:
                     try:
                        edge = (vmap[extra_edge[0]], extra_edge[1], extra_edge[2])
                     except:
                        edge = (extra_edge[0], vmap[extra_edge[1]], extra_edge[2])
                  edge_confs[edge] += consequent_row["confidence"]
                  edge_lifts[edge] += consequent_row["lift"]
                  edge_counts[edge] += 1
      with open("%s_%s_%s_magmascore_%s_%s.csv" % (self.netname, self.min_sup, self.max_size, start, end), 'w') as f:
         f.write("src\ttrg\tlayer\tcount\tconf\tlift\n")
         for edge in edge_lifts:
            line = (self.nodemap[edge[0]], self.nodemap[edge[1]], edge[2], edge_counts[edge], edge_confs[edge], edge_lifts[edge])
            f.write("%s\n" % ('\t'.join([str(x) for x in line])))

   def rules_of_node(self, node): # node 7174 -> 4339
      nodes = list(self.network.vertices())
      egonodes = set([int(node),]) 
      egonodes |= set([int(n1) for n1 in nodes[node].all_neighbors()])
      egonodes |= set([int(n2) for n1 in nodes[node].all_neighbors() for n2 in nodes[int(n1)].all_neighbors()])
      egofilter = np.zeros(len(nodes))
      egofilter[list(egonodes)] = 1
      ego = gt.GraphView(self.network, directed = True, vfilt = egofilter)
      for a in set(self.rules["antecedent"]):
         a = self.patterns_by_id[a]
         break
