import glob, sys, os, Orange, subprocess
import numpy as np
import pandas as pd
import networkx as nx
from libraries.MELL.MELL import MELL_model
from libraries.magma import Magma
from scipy.stats import entropy
from itertools import combinations
from collections import defaultdict
from sklearn.naive_bayes import GaussianNB
from sklearn.metrics import roc_curve, auc
from sklearn.ensemble import RandomForestClassifier

def _drop_existing_edges(infile, scores, directed = False):
   edges = pd.read_csv(infile, sep = "\t", header = None, names = ("src", "trg", "layer"))
   if not directed:
      edges["src_undir"] = edges[["src", "trg"]].min(axis = 1)
      edges["trg_undir"] = edges[["src", "trg"]].max(axis = 1)
      edges[["src", "trg"]] = edges[["src_undir", "trg_undir"]]
      edges = edges.drop(["src_undir", "trg_undir"], axis = 1)
   scores = scores.merge(edges, on = ["src", "trg", "layer"], how = "left", indicator = True)
   scores = scores[scores["_merge"] == "left_only"].drop("_merge", axis = 1)
   return scores

def magma(infile, min_sup, max_size, directed = False, predict_nodes = True, memory = 6, run = True, rule_bounds = None, drop_existing = True):
   m = Magma(infile, min_sup, max_size, directed = directed, memory = memory)
   if run:
      m.run()
   else:
      try:
         m.rules = pd.read_csv("%s.rules" % infile, sep = "\t")
         m.read_patterns()
      except:
         sys.stderr.write("No .rules file found. You should run full Magma (set 'run' parameter to True) for this parameter combination first. Aborting.\n")
         return None
   if rule_bounds is None:
      rule_bounds = (m.rules["antecedent"].min() - 1, m.rules["antecedent"].max() + 1)
   if len(rule_bounds) == 2 and rule_bounds[0] <= rule_bounds[1]:
      m.apply_rules(start = rule_bounds[0], end = rule_bounds[1])
   else:
      sys.stderr.write("Invalid rule bounds. Skipping score step.\n")
      return None
   scores = pd.read_csv("%s_%s_%s_magmascore_%s_%s.csv" % (infile, min_sup, max_size, rule_bounds[0], rule_bounds[1]), sep = "\t")
   if not predict_nodes:
      scores = scores.loc[(scores["src"] > 0) & (scores["trg"] > 0)]
   if not directed:
      scores["src_undir"] = scores[["src", "trg"]].min(axis = 1)
      scores["trg_undir"] = scores[["src", "trg"]].max(axis = 1)
      scores[["src", "trg"]] = scores[["src_undir", "trg_undir"]]
      scores = scores.drop(["src_undir", "trg_undir"], axis = 1)
   scores = scores[scores["src"] != scores["trg"]]
   scores = scores.groupby(by = ["src", "trg", "layer"])["conf"].sum().reset_index()
   scores = scores.rename(columns = {"conf": "magma"})
   if drop_existing:
      scores = _drop_existing_edges(infile, scores, directed = directed)
   scores.to_csv("%s_%s_%s_magmascore_%s_%s.csv" % (infile, min_sup, max_size, rule_bounds[0], rule_bounds[1]), sep = "\t", index = False)
   return scores

def sharma(infile, directed = False, drop_existing = True):
   def _common_edges(edges, layer1, layer2):
      edges_12 = edges[(edges["layer"] == layer1) | (edges["layer"] == layer2)]
      edges_12 = edges_12[edges_12.duplicated(subset = ["src", "trg"])]
      return len(set(zip(edges_12["src"], edges_12["trg"])))
   
   edges = pd.read_csv(infile, sep = "\t", header = None, names = ("src", "trg", "layer"))
   nodes = set(edges["src"]) | set(edges["trg"])
   nodepairs = set(combinations(nodes, 2))
   layers = set(edges["layer"])
   layerpairs = set(combinations(layers, 2))
   edges_layer = {}
   for layer in layers:
      edges_layer[layer] = edges[edges["layer"] == layer].shape[0] / len(nodepairs)
   target_predictor_weight = {}
   for layerpair in layerpairs:
      layerpair_p = _common_edges(edges, layerpair[0], layerpair[1]) / len(nodepairs)
      target_predictor_weight[(layerpair[0], layerpair[1])] = layerpair_p / edges_layer[layerpair[1]]
      target_predictor_weight[(layerpair[1], layerpair[0])] = layerpair_p / edges_layer[layerpair[0]]
   edges = set(zip(edges["src"], edges["trg"], edges["layer"]))
   scores = defaultdict(float)
   for test_edge in nodepairs:
      for layer_pair in target_predictor_weight:
         if (test_edge[0], test_edge[1], layer_pair[1]) in edges and not (test_edge[0], test_edge[1], layer_pair[0]) in edges:
            scores[(test_edge[0], test_edge[1], layer_pair[0])] += target_predictor_weight[layer_pair]
         if directed and (test_edge[1], test_edge[0], layer_pair[1]) in edges and not (test_edge[1], test_edge[0], layer_pair[0]) in edges:
            scores[(test_edge[1], test_edge[0], layer_pair[0])] += target_predictor_weight[layer_pair]
   scores = pd.DataFrame.from_dict(scores, orient = "index", columns = ("sharma",))
   scores.index = pd.MultiIndex.from_tuples(scores.index, names = ("src", "trg", "layer"))
   scores = scores[scores["sharma"] > 0].reset_index()
   if drop_existing:
      scores = _drop_existing_edges(infile, scores, directed = directed)
   return scores

def pujari(infile, directed = False, drop_existing = True):
   def _scores_per_layer(G_all, layer, directed):
      if directed:
         G = nx.DiGraph()
      else:
         G = nx.Graph()
      G.add_edges_from([(e[0], e[1]) for e in G_all.edges(data = True) if e[2]["layer"] == layer])
      G_scores = defaultdict(lambda : defaultdict(float))
      edges = set(G.edges)
      if directed:
         for u, v in edges | set(nx.non_edges(G)):
            u_neigh = set(G.successors(u)) | set(G.predecessors(u))
            v_neigh = set(G.successors(v)) | set(G.predecessors(v))
            neigh_intersection = list(u_neigh & v_neigh)
            G_scores[(u, v)]["cn"] = len(neigh_intersection)
            G_scores[(u, v)]["jc"] = G_scores[(u, v)]["cn"] / len(u_neigh | v_neigh)
            G_scores[(u, v)]["pa"] = G.out_degree(u) * G.in_degree(v)
            G_scores[(u, v)]["aa"] = sum(1 / np.log(G.degree(w)) for w in neigh_intersection)
            G_scores[(u, v)]["ra"] = sum(1 / G.degree(w) for w in neigh_intersection)
            G_scores[(u, v)]["target"] = 1 if (u, v) in edges else 0
      else:
         for u, v in edges | set(nx.non_edges(G)):
            neigh_intersection = list(nx.common_neighbors(G, u, v))
            G_scores[(min(u, v), max(u, v))]["cn"] = len(neigh_intersection)
            G_scores[(min(u, v), max(u, v))]["jc"] = G_scores[(min(u, v), max(u, v))]["cn"] / len(set(G[u]) | set(G[v]))
            G_scores[(min(u, v), max(u, v))]["pa"] = G.degree(u) * G.degree(v)
            G_scores[(min(u, v), max(u, v))]["aa"] = sum(1 / np.log(G.degree(w)) for w in neigh_intersection)
            G_scores[(min(u, v), max(u, v))]["ra"] = sum(1 / G.degree(w) for w in neigh_intersection)
            G_scores[(min(u, v), max(u, v))]["target"] = 1 if (u, v) in edges else 0
      return G_scores

   def _score_multi_layer(G_train):
      G_scores = defaultdict(lambda : defaultdict(float))
      nodepairs = set([pair for layer in G_train for pair in G_train[layer].keys()])
      for k in nodepairs:
         for attribute in ("cn", "jc", "pa", "aa", "ra"):
            total = sum(G_train[layer][k][attribute] for layer in G_train)
            if total > 0:
               G_scores[k]["avg_%s" % attribute] = total / len(G_train)
               G_scores[k]["ent_%s" % attribute] = entropy([G_train[layer][k][attribute] / total for layer in G_train])
      return pd.DataFrame().from_dict(G_scores, orient = "index").reset_index().fillna(0)

   if directed:
      G_all = nx.read_edgelist(infile, delimiter = "\t", data = [("layer", int),], create_using = nx.MultiDiGraph(), nodetype = int)
   else:
      G_all = nx.read_edgelist(infile, delimiter = "\t", data = [("layer", int),], create_using = nx.MultiGraph(), nodetype = int)
   layers = set([x[2]["layer"] for x in G_all.edges(data = True)])
   G_train = {}
   for layer in layers:
      G_train[layer] = _scores_per_layer(G_all, layer, directed)
   G_train["ml"] = _score_multi_layer(G_train)
   for layer in layers:
      G_train[layer] = pd.DataFrame().from_dict(G_train[layer], orient = "index").reset_index().dropna().merge(G_train["ml"], on = ["level_0", "level_1"])
      G_train[layer]["layer"] = layer
   G_train = pd.concat([G_train[layer] for layer in layers])
   netname = infile.split('/')[-1]
   trainfile = f"pujari_train_{netname}.csv"
   with open(trainfile, 'w') as f:
      f.write("%s\n" % '\t'.join(G_train.drop(["level_0", "level_1", "layer"], axis = 1).columns))
      f.write("c\tc\tc\tc\tc\td\tc\tc\tc\tc\tc\tc\tc\tc\tc\tc\n")
      f.write("\t\t\t\t\tclass\t\t\t\t\t\t\t\t\t\t\n")
      G_train.drop(["level_0", "level_1", "layer"], axis = 1).to_csv(f, sep = "\t", index = False, header = False)
   G_train_orange = Orange.data.Table.from_file(trainfile)
   model = Orange.classification.TreeLearner(sufficient_majority = 0.99)
   model = model(G_train_orange)
   os.remove(trainfile)
   scores = G_train.copy().drop("target", axis = 1)
   scores["pujari"] = [model(scores.iloc[i].drop(["level_0", "level_1", "layer"]).values, 1)[0] for i in range(scores.shape[0])]
   scores = scores[["level_0", "level_1", "layer", "pujari"]].rename(columns = {"level_0": "src", "level_1": "trg"})
   scores = scores[scores["pujari"] > 0]
   if drop_existing:
      scores = _drop_existing_edges(infile, scores, directed = directed)
   return scores

def jalili(infile, directed = False, drop_existing = True):
   def _metagraph(G_all, layer, directed, maxnode, infile):
      if directed:
         G = nx.DiGraph()
      else:
         G = nx.Graph()
      netname = infile.split('/')[-1]
      G.add_nodes_from(G_all.nodes)
      G.add_edges_from([(e[0], e[1]) for e in G_all.edges(data = True) if e[2]["layer"] == layer])
      layerfile = f"{netname}_l{layer}"
      nx.write_edgelist(G, layerfile, delimiter = "\t", data = False)
      bash_command = f"./libraries/infomap {layerfile} . -N10"
      myproc = subprocess.Popen(bash_command.split(), stdout = subprocess.PIPE, stderr = subprocess.PIPE)
      myproc.communicate()
      comm_node = defaultdict(list)
      node_comm = {}
      with open(f"{layerfile}.tree", 'r') as f:
         for line in f:
            if line[0] != '#':
               fields = line.strip().split()
               node = int(fields[-1])
               if node in G.nodes:
                  comm = int(fields[0].strip().split(':')[0])
                  node_comm[node] = comm
                  comm_node[comm].append(node)
      maxnode_found = max(node_comm.keys())
      for n in training_set_nodes - set(node_comm):
         comm_id = max(node_comm.values()) + 1
         node_comm[n] = comm_id
         comm_node[comm_id].append(n)
      for comm in comm_node:
         for i in range(1, len(comm_node[comm])):
            G = nx.contracted_nodes(G, comm_node[comm][0], comm_node[comm][i], self_loops = False)
      G = nx.relabel_nodes(G, {n: node_comm[n] for n in G.nodes})
      os.remove(layerfile)
      os.remove(f"{layerfile}.tree")
      return G, node_comm, comm_node
   
   training_set = pd.read_csv(infile, sep = "\t", header = None, names = ("src", "trg", "layer"), dtype = {"layer": str})
   training_set_nodes = set(training_set["src"]) | set(training_set["trg"])
   if directed:
      G_all = nx.read_edgelist(infile, delimiter = "\t", data = [("layer", int),], create_using = nx.MultiDiGraph(), nodetype = int)
      G_all_meta = nx.DiGraph()
   else:
      G_all = nx.read_edgelist(infile, delimiter = "\t", data = [("layer", int),], create_using = nx.MultiGraph(), nodetype = int)
      G_all_meta = nx.Graph()
      training_set["src_undir"] = training_set[["src", "trg"]].min(axis = 1)
      training_set["trg_undir"] = training_set[["src", "trg"]].max(axis = 1)
      training_set[["src", "trg"]] = training_set[["src_undir", "trg_undir"]]
      training_set = training_set.drop(["src_undir", "trg_undir"], axis = 1)
   training_set = pd.MultiIndex.from_frame(training_set)
   layers = set([x[2]["layer"] for x in G_all.edges(data = True)])
   node_comms = {}
   comm_nodes = {}
   meta_nodes = []
   for layer in layers:
      G, node_comm, comm_node = _metagraph(G_all, layer, directed, training_set_nodes, infile)
      node_comms[layer] = node_comm
      comm_nodes[layer] = comm_node
      G_all_meta.add_edges_from([("%s_%s" % (e[0], layer), "%s_%s" % (e[1], layer)) for e in G.edges])
      meta_nodes.extend(["%s_%s" % (metanode, layer) for metanode in comm_nodes[layer]])
   for i in range(len(meta_nodes) - 1):
      i_parts = meta_nodes[i].split('_')
      comm_i = int(i_parts[0])
      layer_i = int(i_parts[1])
      for j in range(i + 1, len(meta_nodes)):
         j_parts = meta_nodes[j].split('_')
         comm_j = int(j_parts[0])
         layer_j = int(j_parts[1])
         if len(set(comm_nodes[layer_i][comm_i]) & set(comm_nodes[layer_j][comm_j])) > 0:
            G_all_meta.add_edge(meta_nodes[i], meta_nodes[j])
            if directed:
               G_all_meta.add_edge(meta_nodes[j], meta_nodes[i])
   p = {}
   p[1] = nx.to_pandas_adjacency(G_all_meta)
   p[2] = p[1].dot(p[1])
   p[3] = p[2]
   np.fill_diagonal(p[3].values, 0)
   p[3] = p[3].dot(p[1])
   p[1] = p[1].to_dict()
   p[2] = p[2].to_dict()
   p[3] = p[3].to_dict()
   nodes = sorted(list(set(G_all.nodes)))
   scores = []
   if directed:
      for i in range(len(nodes)):
         cis = {"%s_%s" % (node_comms[layer][nodes[i]], layer) for layer in node_comms}
         for j in range(len(nodes)):
            if i != j:
               cjs = {"%s_%s" % (node_comms[layer][nodes[j]], layer) for layer in node_comms}
               layer_feature = defaultdict(lambda : defaultdict(int))
               for ci in cis:
                  for cj in cjs:
                     for l in (1, 2, 3):
                        layer_feature[ci.split('_')[1]][l] += p[l][ci][cj]
                        layer_feature[cj.split('_')[1]][l + 3] += p[l][ci][cj]
               for layer in layer_feature:
                  row = [nodes[i], nodes[j], layer]
                  row.extend([layer_feature[layer][feature] for feature in (1, 2, 3, 4, 5, 6)])
                  scores.append(row)
   else:
      for i in range(len(nodes) - 1):
         cis = {"%s_%s" % (node_comms[layer][nodes[i]], layer) for layer in node_comms}
         for j in range(i + 1, len(nodes)):
            cjs = {"%s_%s" % (node_comms[layer][nodes[j]], layer) for layer in node_comms}
            layer_feature = defaultdict(lambda : defaultdict(int))
            for ci in cis:
               for cj in cjs:
                  for l in (1, 2, 3):
                     layer_feature[ci.split('_')[1]][l] += p[l][ci][cj]
                     layer_feature[cj.split('_')[1]][l + 3] += p[l][ci][cj]
            for layer in layer_feature:
               row = [nodes[i], nodes[j], layer]
               row.extend([layer_feature[layer][feature] for feature in (1, 2, 3, 4, 5, 6)])
               scores.append(row)
   scores = pd.DataFrame(data = scores, columns = ("src", "trg", "layer", 1, 2, 3, 4, 5, 6)).set_index(["src", "trg", "layer"])
   scores["training"] = 0
   scores.loc[training_set, "training"] = 1
   scores["jalili"] = GaussianNB().fit(scores[list(range(1, 7))], scores["training"]).predict_proba(scores[list(range(1, 7))])[:,1]
   scores = scores[scores["jalili"] > 0]
   scores = scores.drop([1,2,3,4,5,6], axis = 1).reset_index()
   scores["layer"] = scores["layer"].astype(int)
   if drop_existing:
      scores = scores[scores["training"] != 1]
   return scores.drop("training", axis = 1)

def hristova(infile, directed = False, drop_existing = True, forest_size = 100):
   def scores_per_layer(G_all, target_layer, layers, directed):
      G_scores = defaultdict(lambda : defaultdict(float))
      for layer in layers:
         if directed:
            G = nx.DiGraph()
         else:
            G = nx.Graph()
         G.add_edges_from([(e[0], e[1]) for e in G_all.edges(data = True) if e[2]["layer"] == layer])
         edges = set(G.edges)
         if layer == target_layer:
            if directed:
               for u, v in edges | set(nx.non_edges(G)):
                  G_scores[(u, v)]["target"] = 1 if (u, v) in edges else 0
            else:
               for u, v in edges | set(nx.non_edges(G)):
                  G_scores[(min(u, v), max(u, v))]["target"] = 1 if (u, v) in edges else 0
         else:
            if directed:
               for u, v in edges | set(nx.non_edges(G)):
                  if u in G.nodes and v in G.nodes:
                     u_neigh = set(G.successors(u)) | set(G.predecessors(u))
                     v_neigh = set(G.successors(v)) | set(G.predecessors(v))
                     neigh_intersection = list(u_neigh & v_neigh)
                     G_scores[(u, v)]["jc%s" % layer] = len(neigh_intersection) / len(u_neigh | v_neigh)
                     G_scores[(u, v)]["aa%s" % layer] = sum(1 / np.log(G.degree(w)) for w in neigh_intersection)
                  else:
                     G_scores[(u, v)]["jc%s" % layer] = 0
                     G_scores[(u, v)]["aa%s" % layer] = 0
            else:
               for u, v in edges | set(nx.non_edges(G)):
                  if u in G.nodes and v in G.nodes:
                     neigh_intersection = list(nx.common_neighbors(G, u, v))
                     G_scores[(min(u, v), max(u, v))]["jc%s" % layer] = len(neigh_intersection) / len(set(G[u]) | set(G[v]))
                     G_scores[(min(u, v), max(u, v))]["aa%s" % layer] = sum(1 / np.log(G.degree(w)) for w in neigh_intersection)
                  else:
                     G_scores[(min(u, v), max(u, v))]["jc%s" % layer] = 0
                     G_scores[(min(u, v), max(u, v))]["aa%s" % layer] = 0
      return G_scores

   def score_multi_layer(G_all, directed):
      G_scores = defaultdict(lambda : defaultdict(float))
      nodepairs = set(combinations(sorted(G_all.nodes), 2))
      if directed:
         for u, v in nodepairs:
            if v in G_all[u]:
               G_scores[(u, v)]["int"] = len(G_all[u][v])
            else:
               G_scores[(u, v)]["int"] = 0
            if u in G_all[v]:
               G_scores[(v, u)]["int"] = len(G_all[v][u])
            else:
               G_scores[(v, u)]["int"] = 0
            u_neigh = set(G_all.successors(u)) | set(G_all.predecessors(u))
            v_neigh = set(G_all.successors(v)) | set(G_all.predecessors(v))
            neigh_intersection = list(u_neigh & v_neigh)
            gn_jc = len(neigh_intersection) / len(u_neigh | v_neigh)
            gn_aa = sum(1 / np.log(G_all.degree(w)) for w in neigh_intersection)
            G_scores[(u, v)]["gn_jc"] = gn_jc
            G_scores[(u, v)]["gn_aa"] = gn_aa
            G_scores[(v, u)]["gn_jc"] = gn_jc
            G_scores[(v, u)]["gn_aa"] = gn_aa
      else:
         for u, v in nodepairs:
            if v in G_all[u]:
               G_scores[(min(u, v), max(u, v))]["int"] = len(G_all[u][v])
            else:
               G_scores[(min(u, v), max(u, v))]["int"] = 0
            neigh_intersection = list(nx.common_neighbors(G_all, u, v))
            G_scores[(min(u, v), max(u, v))]["gn_jc"] = len(neigh_intersection) / len(set(G_all[u]) | set(G_all[v]))
            G_scores[(min(u, v), max(u, v))]["gn_aa"] = sum(1 / np.log(G_all.degree(w)) for w in neigh_intersection)
      return pd.DataFrame().from_dict(G_scores, orient = "index").reset_index().fillna(0)

   if directed:
      G_all = nx.read_edgelist(infile, delimiter = "\t", data = [("layer", int),], create_using = nx.MultiDiGraph(), nodetype = int)
   else:
      G_all = nx.read_edgelist(infile, delimiter = "\t", data = [("layer", int),], create_using = nx.MultiGraph(), nodetype = int)
   layers = set([x[2]["layer"] for x in G_all.edges(data = True)])
   ml_feats = score_multi_layer(G_all, directed)
   scores = []
   for layer in layers:
      df = scores_per_layer(G_all, layer, layers, directed)
      df = pd.DataFrame().from_dict(df, orient = "index").reset_index().dropna(subset = ["target",]).fillna(0).merge(ml_feats, on = ["level_0", "level_1"])
      score = RandomForestClassifier(n_estimators = forest_size).fit(df.drop(["level_0", "level_1", "target"], axis = 1), df["target"]).predict_proba(df.drop(["level_0", "level_1", "target"], axis = 1))
      df["hristova"] = 1 - score[:,0]
      df["layer"] = layer
      scores.append(df[["level_0", "level_1", "layer", "hristova"]])
   scores = pd.concat(scores).rename(columns = {"level_0": "src", "level_1": "trg"})
   scores = scores[scores["hristova"] > 0]
   if drop_existing:
      scores = _drop_existing_edges(infile, scores, directed = directed)
   return scores

def debacco(infile_full, directed = False, drop_existing = True, code_folder = "libraries/debacco", communities = 5):
   infile = infile_full.split('/')[-1]
   df = pd.read_csv(infile_full, sep = "\t", header = None, names = ("src", "trg", "layer"))
   df["weight"] = 1
   layers = len(set(df["layer"]))
   df = pd.pivot_table(data = df, index = ["src", "trg"], columns = "layer", values = "weight").reset_index().fillna(0).astype(int)
   df["preamble"] = 'E'
   df[["preamble", "src", "trg"] + [i + 1 for i in range(layers)]].to_csv("%s/data/%s.dat" % (code_folder, infile), sep = " ", index = False, header = False)
   os.chdir("%s/python/" % code_folder)
   if directed:
      bash_command = "python3 main.py -k=%s -l=%s -a=%s.dat -E=_%s.dat" % (communities, layers, infile, infile)
   else:
      bash_command = "python3 main.py -k=%s -l=%s -a=%s.dat -E=_%s.dat -u=1" % (communities, layers, infile, infile)
   process = subprocess.Popen(bash_command.split(), stdout = subprocess.PIPE)
   output, error = process.communicate()
   os.chdir("../../../")
   names = ["node",] + ["c%s" % (i + 1) for i in range(communities)]
   memberships = pd.read_csv("%s/data/u_K%s_%s.dat" % (code_folder, communities, infile), sep = " ", comment = "#", header = None).iloc[:,:-1]
   memberships.columns = names
   memberships = memberships.set_index("node").unstack().reset_index()
   memberships = memberships[memberships[0] > 0].rename(columns = {"level_0": "c", 0: "w"})
   if directed:
      memberships_trg = pd.read_csv("%s/data/v_K%s_%s.dat" % (code_folder, communities, infile), sep = " ", comment = "#", header = None).iloc[:,:-1]
      memberships_trg.columns = names
      memberships_trg = memberships_trg.set_index("node").unstack().reset_index()
      memberships_trg = memberships_trg[memberships_trg[0] > 0].rename(columns = {"level_0": "c", 0: "w"})
   
   i = 1
   curlayer = 1
   c_c = defaultdict(lambda : defaultdict(float))
   df = []
   with open("%s/data/w_K%s_%s.dat" % (code_folder, communities, infile), 'r') as f:
      for line in f:
         if len(line) < 2:
            _ = pd.DataFrame().from_dict(c_c, orient = "index").unstack().reset_index().rename(columns = {"level_0": "c_src", "level_1": "c_trg", 0: "w"})
            _["layer"] = curlayer
            df.append(_)
            i = 1
            curlayer += 1
            c_c = defaultdict(lambda : defaultdict(float))
         elif line[0] != '#' and line[0] != 'a':
            fields = line.strip().split()
            for j in range(len(fields)):
               c_c["c%s" % i]["c%s" % (j + 1)] = float(fields[j])
            i += 1
   os.remove("%s/data/%s.dat" % (code_folder, infile))
   os.remove("%s/data/u_K%s_%s.dat" % (code_folder, communities, infile))
   os.remove("%s/data/w_K%s_%s.dat" % (code_folder, communities, infile))
   if directed:
      os.remove("%s/data/v_K%s_%s.dat" % (code_folder, communities, infile))
   c_c = pd.concat(df)
   c_c = c_c[c_c["w"] > 0]
   if directed:
      scores = pd.merge(memberships.assign(key = "x"), memberships_trg.assign(key = "x"), on = "key", suffixes = ("_src", "_trg")).merge(c_c, on = ["c_src", "c_trg"]).drop("key", axis = 1)
   else:
      scores = pd.merge(memberships.assign(key = "x"), memberships.assign(key = "x"), on = "key", suffixes = ("_src", "_trg")).merge(c_c, on = ["c_src", "c_trg"]).drop("key", axis = 1)
   scores = scores[scores["node_src"] != scores["node_trg"]]
   scores["debacco"] = scores["w_src"] * scores["w_trg"] * scores["w"]
   if not directed:
      scores["src"] = scores[["node_src", "node_trg"]].min(axis = 1)
      scores["trg"] = scores[["node_src", "node_trg"]].max(axis = 1)
   else:
      scores = scores.rename(columns = {"node_src": "src", "node_trg": "trg"})
   scores = scores.groupby(by = ["src", "trg", "layer"])["debacco"].sum().reset_index()
   if drop_existing:
      scores = _drop_existing_edges(infile_full, scores, directed = directed)
   return scores

def mell(infile, directed = False, drop_existing = True):
   df = pd.read_csv(infile, sep = "\t", header = None, names = ("src", "trg", "layer"))
   df["src"] = df["src"] - 1
   df["trg"] = df["trg"] - 1
   df["layer"] = df["layer"] - 1
   srcs = set(df["src"])
   trgs = set(df["trg"])
   nodes = srcs | trgs
   layers = set(df["layer"])
   if directed:
      scores = [(src, trg, layer) for src in nodes for trg in nodes for layer in layers if src != trg]
   else:
      scores = [(src, trg, layer) for src in nodes for trg in nodes for layer in layers if src < trg]
   scores = pd.DataFrame(data = scores, columns = ("src", "trg", "layer"))
   if drop_existing:
      scores = pd.concat([scores, df])
      scores = scores.drop_duplicates(keep = False)
   train_edges = [[row["layer"], row["src"], row["trg"], 1] for index, row in df.iterrows()]
   model = MELL_model(len(layers), max(nodes) + 1, directed, train_edges, 128, 4, 10, 1, 1)
   model.train(500)
   scores["mell"] = scores.apply(lambda x: model.predict([x["layer"], x["src"], x["trg"], 0]), axis = 1)
   scores[["src", "trg", "layer"]] += 1
   return scores


